﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Diagnostics;
using System.IO;
using System.Text;

//  ProfitUI Quick Raid Buttons tool
//  Copyright (C) 2012 Todd T Knarr <tknarr@silverglass.org> <tknarr@cox.net> <todd.knarr@gmail.com>

//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.

//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.

//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

namespace QuickRaidButtons
{
    /// <summary>
    /// When class radio button is changed, load the button assignments from the internal object.
    /// When a new item from a pull-down box is selected for a button, save the new assignment into
    /// the internal object.
    /// 
    /// When the Save button is clicked, save the button assignments from the internal object into
    /// the QRB storage object and save the QRB document to the storage XML document.
    /// When the Revert button is clicked, load the button assignments from the QRB storage object
    /// into the internal object.
    /// When the Generate UI File button is pressed save (as per the Save button), then open a file
    /// save dialog to pick a file to save to and generate a new UI file through the QRB storage
    /// object. Initially default to the My Documents folder and the standard name of the ProfitUI
    /// quick raid buttons file, and save the new folder and filename if the user changes them.
    /// </summary>
    public partial class ButtonAssignment : Form
    {
        private QRBStorage qrb = null; // Reference to our persistent storage object

        private string currentButtonClass = ""; // The current class radio button selection
        private Dictionary<string, QRBButtonAssignment> buttonAssignments = null; // Our current button assignment state reflecting the GUI state

        private QRBStorage createQRBStorage()
        {
            StringBuilder filePath = new StringBuilder();
            filePath.Append( Environment.GetFolderPath( Environment.SpecialFolder.ApplicationData ) );
            filePath.Append( Path.DirectorySeparatorChar );
            filePath.Append( Properties.Settings.Default.StorageFolder );
            string StoragePath = filePath.ToString();
            if ( !Directory.Exists( StoragePath ) )
                Directory.CreateDirectory( StoragePath );
            filePath.Append( Path.DirectorySeparatorChar );
            filePath.Append( Properties.Settings.Default.StorageFilename );
            string StorageFilename = filePath.ToString();
            Debug.WriteLine( "Storage file name: " + StorageFilename );
            StringBuilder uiPath = new StringBuilder();
            if ( !String.IsNullOrEmpty( Properties.Settings.Default.UIFileFolder ) )
                uiPath.Append( Properties.Settings.Default.UIFileFolder );
            else
            {
                uiPath.Append( Environment.GetFolderPath( Environment.SpecialFolder.MyDocuments ) );
            }
            string StorageUIPath = uiPath.ToString();
            Debug.WriteLine( "UI file path: " + StorageUIPath );
            Debug.WriteLine( "UI file name: " + Properties.Settings.Default.UIFilename );
            QRBStorage qrb = new QRBStorage( StorageFilename, StorageUIPath, Properties.Settings.Default.UIFilename );

            return qrb;
        }

        // Expects currentButtonClass to be set to reflect the class the user's selected.
        // If no entry in the buttonAssignments dictionary exists for that class, an empty one will be created.
        // Populates the spell line combo boxes from what's in the persistent storage object, setting the active
        //     entry based on the button assignment for this class in the buttonAssignments dictionary.
        private void populateButtonAssignments()
        {
            // Update the radio button for the current class
            Control[] panel = Controls.Find( "panel1", true );
            if ( panel.Length != 0 )
            {
                foreach ( Control c in panel[0].Controls )
                {
                    if ( c.Name.StartsWith( "radioButton" ) )
                    {
                        RadioButton rb = c as RadioButton;
                        if ( rb != null )
                        {
                            if ( String.Equals( rb.Text, currentButtonClass ) )
                                rb.Checked = true;
                            else
                                rb.Checked = false;
                        }
                    }
                }
            }

            // Locate the button assignments for the current class
            QRBButtonAssignment currentAssignments = null;
            if ( buttonAssignments.ContainsKey( currentButtonClass ) )
                currentAssignments = buttonAssignments[currentButtonClass];
            if ( currentAssignments == null )
            {
                currentAssignments = new QRBButtonAssignment( currentButtonClass );
                buttonAssignments[currentButtonClass] = currentAssignments;
            }

            // Locate the spell lines for the current class
            QRBSpellClass spellLines = null;
            if ( qrb.SpellClasses.ContainsKey( currentButtonClass ) )
                spellLines = qrb.SpellClasses[currentButtonClass];

            // Populate the pull-down boxes for the 5 buttons we support
            int buttonNumber = 0;
            ComboBox[] cbs = new ComboBox[] { comboBox1, comboBox2, comboBox3, comboBox4, comboBox5 };
            foreach ( ComboBox cb in cbs )
            {
                buttonNumber++;
                Debug.WriteLine( "Button number " + buttonNumber.ToString() );
                int buttonIndex = 0;
                string currentButton = currentAssignments.getButton( buttonNumber );
                cb.Items.Clear();
                // We always have a blank no-assignment possibility
                cb.Items.Add( "" );
                if ( spellLines != null )
                {
                    int i = 1; // Remember that we have that initial blank entry in the combobox
                    foreach ( QRBSpellLine sl in spellLines.SpellLines )
                    {
                        Debug.WriteLine( "Spell line " + i.ToString() + ": " + sl.Linename );
                        cb.Items.Add( sl.Linename );
                        // If the entry we just populated matches the current button assignment, remember
                        // it's index so we can set the drop-down's initial value
                        if ( sl.Linename == currentButton )
                            buttonIndex = i;
                        i++;
                    }
                }

                Debug.WriteLine( "Active spell line: " + buttonIndex.ToString() );
                // Set the initial item in the drop-down to the current selection, or to the
                // blank line if we didn't have a selection.
                cb.SelectedIndex = buttonIndex;
                //cb.SelectedText = currentButton;
            }
        }

        private void ButtonAssignment_FormClosing( object sender, FormClosingEventArgs e )
        {
            Debug.WriteLine( "Button assignment form closing" );
            // Save as part of closing
            qrb.ButtonAssignments = buttonAssignments;
            qrb.saveDocument();

            // Update our settings
            if ( qrb.UIPath != Properties.Settings.Default.UIFileFolder )
                Properties.Settings.Default.UIFileFolder = qrb.UIPath;
            if ( qrb.UIFilename != Properties.Settings.Default.UIFilename )
                Properties.Settings.Default.UIFilename = qrb.UIFilename;
            Properties.Settings.Default.InitialClass = currentButtonClass;
            Properties.Settings.Default.Save();
        }

        public ButtonAssignment()
        {
            InitializeComponent();

            qrb = createQRBStorage();
            buttonAssignments = qrb.CloneButtonAssignments();
            if ( !String.IsNullOrEmpty( Properties.Settings.Default.InitialClass ) )
                currentButtonClass = Properties.Settings.Default.InitialClass;
            else
                currentButtonClass = "Guardian";
            Debug.WriteLine( "Initial class selected: " + currentButtonClass );

            populateButtonAssignments();
        }

        private void exitButton_Click( object sender, EventArgs e )
        {
            qrb.ButtonAssignments = buttonAssignments;
            qrb.saveDocument();
            Application.Exit();
        }

        private void generateButton_Click( object sender, EventArgs e )
        {
            Debug.WriteLine( "Generate UI code button clicked" );
            qrb.ButtonAssignments = buttonAssignments;
            buttonAssignments = qrb.CloneButtonAssignments();
            qrb.saveDocument();

            // Open a file save dialog, setting the path and filename based on the UIPath and UIFilename in the qrb object
            SaveFileDialog dlg = new SaveFileDialog();
            dlg.InitialDirectory = qrb.UIPath;
            dlg.FileName = qrb.UIFilename;
            dlg.DefaultExt = ".txt";
            dlg.AddExtension = true;
            dlg.CheckPathExists = true;
            dlg.OverwritePrompt = true;
            dlg.ValidateNames = true;
            dlg.Filter = "Text files|*.txt|XML files|*.xml|All files|*.*";
            DialogResult result = dlg.ShowDialog( this );
            if ( result == DialogResult.OK )
            {
                string fp = dlg.FileName;
                Debug.Write( "UI code selected file: " + fp );
                string fd = Path.GetDirectoryName( fp );
                Debug.Write( "UI code selected directory: " + fd );
                string fn = Path.GetFileName( fp );
                Debug.Write( "UI code selected filename: " + fn );
                // Call the qrb object's saveUICode() method to write the file selected in the file save dialog
                qrb.saveUICode( fd, fn );
            }
        }

        private void saveButton_Click( object sender, EventArgs e )
        {
            qrb.ButtonAssignments = buttonAssignments;
            buttonAssignments = qrb.CloneButtonAssignments();
            qrb.saveDocument();
        }

        private void revertButton_Click( object sender, EventArgs e )
        {
            buttonAssignments = qrb.CloneButtonAssignments();
            populateButtonAssignments();
        }

        private void editSpellLinesButton_Click( object sender, EventArgs e )
        {
            using ( SpellLines sl_window = new SpellLines( qrb, currentButtonClass ) )
            {
                sl_window.ShowDialog( this );
                populateButtonAssignments();
            }
        }

        private void classSelection_Click( object sender, EventArgs e )
        {
            RadioButton btn = sender as RadioButton;
            currentButtonClass = btn.Text;
            Debug.WriteLine( "New class selected: " + currentButtonClass );
            populateButtonAssignments();
        }

        private void buttonSpellLine_SelectionChanged( object sender, EventArgs e )
        {
            ComboBox spellLineName = sender as ComboBox;
            int button_number = int.Parse( spellLineName.Tag.ToString() );
            string line_name = spellLineName.Text;

            if ( buttonAssignments.ContainsKey( currentButtonClass ) )
                buttonAssignments[currentButtonClass].setButton( button_number, line_name );
        }

        private void aboutButton_Click( object sender, EventArgs e )
        {
            AboutBox a = new AboutBox();
            a.ShowDialog( this );
        }
    }
}
