﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Xml;
using System.Xml.Schema;
using System.IO;
using System.Diagnostics;
using System.Text;

//  ProfitUI Quick Raid Buttons tool
//  Copyright (C) 2012 Todd T Knarr <tknarr@silverglass.org> <tknarr@cox.net> <todd.knarr@gmail.com>

//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.

//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.

//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

namespace QuickRaidButtons
{
    public class QRBSpell : IComparable, ICloneable
    {
        // Data members
        public int Level { get; set; }
        public int Sequence { get; set; }
        public string Text { get; set; }

        public QRBSpell()
        {
            Level = 0;
            Sequence = 0;
            Text = "";
        }

        public QRBSpell( int l, string t )
        {
            Level = l;
            Sequence = 0;
            Text = t;
        }

        public QRBSpell( int l, int s, string t )
        {
            Level = l;
            Sequence = s;
            Text = t;
        }

        public object Clone()
        {
            return new QRBSpell( Level, Sequence, Text );
        }

        public int CompareTo( object obj )
        {
            QRBSpell right = obj as QRBSpell;
            if ( right != null )
            {
                int result = Level.CompareTo( right.Level );
                if ( result == 0 )
                    return Sequence.CompareTo( right.Sequence );
                else
                    return result;
            }
            else
                throw new ArgumentException( "Object is not a QRBSpell" );
        }
    }

    public class QRBSpellLine : IComparable, ICloneable
    {
        // Data members
        public string Linename { get; set; }
        public string Tooltip { get; set; }
        public List<QRBSpell> Spells { get; set; }

        public QRBSpellLine()
        {
            Linename = "";
            Tooltip = "";
            Spells = new List<QRBSpell>();
        }

        public QRBSpellLine( string ln, string tt )
        {
            Linename = ln;
            Tooltip = tt;
            Spells = new List<QRBSpell>();
        }

        public bool hasSpells()
        {
            foreach ( QRBSpell s in Spells )
                if ( s.Text != "" )
                    return true;
            return false;
        }

        public void addSpell( QRBSpell s )
        {
            if ( Spells.Count > 0 )
                s.Sequence = Spells[Spells.Count - 1].Sequence + 1;
            else
                s.Sequence = 1;
            Spells.Add( s );
            return;
        }

        public void addSpell( int l, string t )
        {
            addSpell( new QRBSpell( l, t ) );
            return;
        }

        public void resequenceSpells()
        {
            Spells.Sort();
            int sequence = 1;
            foreach ( QRBSpell s in Spells )
            {
                s.Sequence = sequence;
                sequence++;
            }
            return;
        }

        public object Clone()
        {
            QRBSpellLine result = new QRBSpellLine( Linename, Tooltip );
            foreach ( QRBSpell s in Spells )
                result.Spells.Add( s.Clone() as QRBSpell );
            return result;
        }

        public int CompareTo( object obj )
        {
            QRBSpellLine right = obj as QRBSpellLine;
            if ( right != null )
                return Linename.CompareTo( right.Linename );
            else
                throw new ArgumentException( "Object is not a QRBSpellLine" );
        }
    }

    public class QRBSpellClass : IComparable, ICloneable
    {
        // Data members
        public string Classname { get; set; }
        public List<QRBSpellLine> SpellLines { get; set; }

        public QRBSpellClass()
        {
            Classname = "";
            SpellLines = new List<QRBSpellLine>();
        }

        public QRBSpellClass( string cn )
        {
            Classname = cn;
            SpellLines = new List<QRBSpellLine>();
        }

        public bool hasSpellLines()
        {
            foreach ( QRBSpellLine sl in SpellLines )
                if ( sl.hasSpells() )
                    return true;
            return false;
        }

        public QRBSpellLine findSpellLine( string s )
        {
            foreach ( QRBSpellLine sl in SpellLines )
            {
                if ( sl.Linename.Equals( s ) )
                    return sl;
            }
            return null;
        }

        public QRBSpellLine addSpellLine( string s, string t = "" )
        {
            foreach ( QRBSpellLine sl in SpellLines )
            {
                if ( sl.Linename.Equals( s ) )
                    return sl;
            }
            QRBSpellLine n = new QRBSpellLine( s, t );
            SpellLines.Add( n );
            return n;
        }

        public void removeSpellLine( string s )
        {
            for ( int i = 0; i < SpellLines.Count; i++ )
            {
                if ( SpellLines[i].Linename.Equals( s ) )
                {
                    SpellLines.RemoveAt( i );
                    i--;
                }
            }
            return;
        }

        public object Clone()
        {
            QRBSpellClass result = new QRBSpellClass( Classname );
            foreach ( QRBSpellLine sl in SpellLines )
                result.SpellLines.Add( sl.Clone() as QRBSpellLine );
            return result;
        }

        public int CompareTo( object obj )
        {
            QRBSpellClass right = obj as QRBSpellClass;
            if ( right != null )
                return Classname.CompareTo( right.Classname );
            else
                throw new ArgumentException( "Object is not a QRBSpellClass" );
        }
    }

    public class QRBButtonAssignment : IComparable, ICloneable
    {
        // Data members
        public string Classname { get; set; }
        private string[] button;

        public QRBButtonAssignment()
        {
            Classname = "";
            button = new string[5];
        }

        public QRBButtonAssignment( string cn )
        {
            Classname = cn;
            button = new string[5];
        }

        public string getButton( int n )
        {
            if ( ( n >= 1 ) && ( n <= 5 ) )
                return button[n - 1];
            else
                return "";
        }

        public void setButton( int n, string ln )
        {
            if ( ( n >= 1 ) && ( n <= 5 ) )
                button[n - 1] = ln;
            return;
        }

        public int buttonCount()
        {
            int result = 0;
            for ( int i = 0; i < 5; i++ )
                if ( !string.IsNullOrEmpty( button[i] ) )
                    result++;
            return result;
        }

        public object Clone()
        {
            QRBButtonAssignment result = new QRBButtonAssignment( Classname );
            for ( int i = 0; i < 5; i++ )
                result.button[i] = button[i];
            return result;
        }

        public int CompareTo( object obj )
        {
            QRBButtonAssignment right = obj as QRBButtonAssignment;
            if ( right != null )
                return Classname.CompareTo( right.Classname );
            else
                throw new ArgumentException( "Object is not a QRBButtonAssignment" );
        }
    }

    public class QRBStorage : ICloneable
    {
        // Data members
        public string StorageFilename { get; private set; }
        public string UIPath { get; set; }
        public string UIFilename { get; set; }
        public Dictionary<string, QRBSpellClass> SpellClasses { get; set; }
        public Dictionary<string, QRBButtonAssignment> ButtonAssignments { get; set; }

        public QRBStorage()
        {
            SpellClasses = new Dictionary<string, QRBSpellClass>();
            ButtonAssignments = new Dictionary<string, QRBButtonAssignment>();
            StorageFilename = "";
            UIPath = "";
            UIFilename = "";
        }

        public QRBStorage( string fileName, string uiPath, string uiFilename )
        {
            SpellClasses = new Dictionary<string, QRBSpellClass>();
            ButtonAssignments = new Dictionary<string, QRBButtonAssignment>();
            StorageFilename = fileName;
            UIPath = uiPath;
            UIFilename = uiFilename;

            loadDocument();
        }

        public object Clone()
        {
            QRBStorage result = new QRBStorage();
            result.StorageFilename = StorageFilename.Clone() as string;
            result.UIPath = UIPath.Clone() as string;
            result.UIFilename = UIFilename.Clone() as string;
            result.SpellClasses = CloneSpellClasses();
            result.ButtonAssignments = CloneButtonAssignments();
            return result;
        }

        private void loadSpells( XmlElement root )
        {
            foreach ( XmlNode node in root.ChildNodes )
            {
                if ( node.LocalName.Equals( "SpellClass" ) )
                {
                    XmlElement elem = (XmlElement) node;
                    string classname = elem.GetAttribute( "class" );
                    if ( !SpellClasses.ContainsKey( classname ) )
                    {
                        QRBSpellClass sc = new QRBSpellClass( classname );

                        foreach ( XmlNode node2 in elem.ChildNodes )
                        {
                            if ( node2.LocalName.Equals( "SpellLine" ) )
                            {
                                int sequence = 1;

                                XmlElement elem2 = (XmlElement) node2;
                                string linename = elem2.GetAttribute( "name" );
                                string tooltip = "";
                                List<QRBSpell> sps = new List<QRBSpell>();
                                foreach ( XmlNode node3 in elem2.ChildNodes )
                                {
                                    if ( node3.LocalName.Equals( "ToolTip" ) )
                                    {
                                        XmlElement elem3 = (XmlElement) node3;
                                        tooltip = elem3.InnerText;
                                    }
                                    else if ( node3.LocalName.Equals( "Spell" ) )
                                    {
                                        int level = -1;
                                        string text = "";
                                        foreach ( XmlNode node4 in node3.ChildNodes )
                                        {
                                            if ( node4.LocalName.Equals( "Level" ) )
                                            {
                                                XmlElement elem4 = (XmlElement) node4;
                                                level = int.Parse( elem4.InnerText );
                                            }
                                            else if ( node4.LocalName.Equals( "Text" ) )
                                            {
                                                XmlElement elem4 = (XmlElement) node4;
                                                text = elem4.InnerText;
                                            }
                                        }
                                        if ( level >= 0 )
                                        {
                                            QRBSpell sp = new QRBSpell( level, sequence, text );
                                            sps.Add( sp );
                                            sequence++;
                                        }
                                    }
                                }

                                QRBSpellLine sl = new QRBSpellLine( linename, tooltip );
                                if ( sps.Count > 0 )
                                    sl.Spells = sps;
                                sc.SpellLines.Add( sl );
                            }
                        }

                        if ( sc.hasSpellLines() )
                            SpellClasses.Add( classname, sc );
                    }
                }
            }
        }

        private void loadButtonAssignments( XmlElement root )
        {
            foreach ( XmlNode node in root.ChildNodes )
            {
                if ( node.LocalName.Equals( "ButtonClass" ) )
                {
                    XmlElement elem = (XmlElement) node;
                    string classname = elem.GetAttribute( "class" );
                    if ( !ButtonAssignments.ContainsKey( classname ) )
                    {
                        QRBButtonAssignment ba = new QRBButtonAssignment( classname );

                        foreach ( XmlNode node2 in elem.ChildNodes )
                        {
                            if ( node2.LocalName.Equals( "Button" ) )
                            {
                                XmlElement elem2 = (XmlElement) node2;
                                int btn_number = int.Parse( elem2.GetAttribute( "number" ) );
                                string spelllinename = "";
                                foreach ( XmlNode node3 in elem2.ChildNodes )
                                {
                                    if ( node3.LocalName.Equals( "SpellLineName" ) )
                                    {
                                        XmlElement elem3 = (XmlElement) node3;
                                        spelllinename = elem3.InnerText;
                                    }
                                }
                                if ( spelllinename != "" )
                                    ba.setButton( btn_number, spelllinename );
                            }
                        }

                        if ( ba.buttonCount() > 0 )
                            ButtonAssignments.Add( classname, ba );
                    }
                }
            }
        }

        private void loadDocument()
        {
            SpellClasses.Clear();
            ButtonAssignments.Clear();

            if ( !string.IsNullOrEmpty( StorageFilename ) && File.Exists( StorageFilename ) )
            {
                // Create an XmlReader and use it to create an XmlDocument read from the storage file
                XmlReaderSettings settings = new XmlReaderSettings();
                settings.IgnoreWhitespace = true;
                settings.IgnoreComments = true;
                settings.CloseInput = true;
                settings.Schemas.Add( "http://xsd.silverglass.org/QuickRaidButtons/QRBStorage.xsd", "QRBStorage.xsd" );
                settings.ValidationType = ValidationType.Schema;
                using ( XmlReader reader = XmlReader.Create( StorageFilename, settings ) )
                {
                    XmlDocument doc = new XmlDocument();
                    doc.Load( reader );
                    XmlElement root = doc.DocumentElement;
                    foreach ( XmlNode node in root.ChildNodes )
                    {
                        if ( node.GetType() == typeof( XmlElement ) )
                        {
                            XmlElement elem = (XmlElement) node;
                            if ( elem.LocalName.Equals( "Spells" ) )
                                loadSpells( elem );
                            else if ( node.LocalName.Equals( "ButtonAssignments" ) )
                                loadButtonAssignments( elem );
                        }
                    }
                }
            }
        }

        public void saveDocument()
        {
            if ( StorageFilename != "" )
            {
                XmlWriterSettings settings = new XmlWriterSettings();
                settings.Indent = true;
                settings.CloseOutput = true;
                using ( XmlWriter writer = XmlWriter.Create( StorageFilename, settings ) )
                {
                    writer.WriteStartDocument();
                    writer.WriteStartElement( "QuickRaidButtons", "http://xsd.silverglass.org/QuickRaidButtons/QRBStorage.xsd" );

                    writer.WriteStartElement( "Spells" );
                    foreach ( KeyValuePair<string, QRBSpellClass> kv in SpellClasses )
                    {
                        QRBSpellClass sc = kv.Value;

                        writer.WriteStartElement( "SpellClass" );
                        writer.WriteAttributeString( "class", sc.Classname );
                        if ( sc.hasSpellLines() )
                        {
                            foreach ( QRBSpellLine sl in sc.SpellLines )
                            {
                                writer.WriteStartElement( "SpellLine" );
                                writer.WriteAttributeString( "name", sl.Linename );
                                if ( !string.IsNullOrEmpty( sl.Tooltip ) )
                                {
                                    writer.WriteElementString( "ToolTip", sl.Tooltip );
                                }
                                if ( sl.hasSpells() )
                                {
                                    foreach ( QRBSpell sp in sl.Spells )
                                    {
                                        writer.WriteStartElement( "Spell" );
                                        writer.WriteElementString( "Level", sp.Level.ToString() );
                                        writer.WriteElementString( "Text", sp.Text );
                                        writer.WriteEndElement();
                                    }
                                }
                                writer.WriteEndElement(); // SpellLine
                            }
                        }
                        writer.WriteEndElement(); // SpellClass
                    }
                    writer.WriteEndElement(); // Spells

                    writer.WriteStartElement( "ButtonAssignments" );
                    foreach ( KeyValuePair<string, QRBButtonAssignment> kv in ButtonAssignments )
                    {
                        QRBButtonAssignment ba = kv.Value;

                        if ( ba.buttonCount() > 0 )
                        {
                            writer.WriteStartElement( "ButtonClass" );
                            writer.WriteAttributeString( "class", ba.Classname );
                            for ( int i = 1; i <= 5; i++ )
                            {
                                string s = ba.getButton( i );
                                if ( !string.IsNullOrEmpty( s ) )
                                {
                                    writer.WriteStartElement( "Button" );
                                    writer.WriteAttributeString( "number", i.ToString() );
                                    writer.WriteElementString( "SpellLineName", s );
                                    writer.WriteEndElement();
                                }
                            }
                            writer.WriteEndElement(); // ButtonClass
                        }
                    }
                    writer.WriteEndElement(); // ButtonAssignments

                    writer.WriteEndElement(); // QuickRaidButtons
                    writer.WriteEndDocument();
                    writer.Close();
                }
            }
        }

        public Dictionary<string, QRBSpellClass> CloneSpellClasses()
        {
            Dictionary<string, QRBSpellClass> result = new Dictionary<string, QRBSpellClass>();
            foreach ( KeyValuePair<string, QRBSpellClass> kv in SpellClasses )
                result.Add( kv.Key, kv.Value.Clone() as QRBSpellClass );
            return result;
        }

        public Dictionary<string, QRBButtonAssignment> CloneButtonAssignments()
        {
            Dictionary<string, QRBButtonAssignment> result = new Dictionary<string, QRBButtonAssignment>();
            foreach ( KeyValuePair<string, QRBButtonAssignment> kv in ButtonAssignments )
                result.Add( kv.Key, kv.Value.Clone() as QRBButtonAssignment );
            return result;
        }

        private string doUISubstitutions( string s )
        {
            // Trim leading and trailing whitespace and do percent-variable substitution
            // %T means target, %I means target's target (implied target)
            string ours = s.Trim();
            ours = ours.Replace( "%T", "Parent.Target" );
            ours = ours.Replace( "%t", "Parent.Target" );
            ours = ours.Replace( "%I", "Parent.Target.Target" );
            ours = ours.Replace( "%i", "Parent.Target.Target" );

            // Encode special characters that need it
            // Currently this is just single- and double-quote characters
            StringBuilder sb = new StringBuilder();
            foreach ( char c in ours )
            {
                if ( c == '\"' )
                    sb.Append( "&quot;" );
                else if ( c == '\'' )
                    sb.Append( "&apos;" );
                else
                    sb.Append( c );
            }
            return sb.ToString();
        }

        public void saveUICode( string pathName, string fileName )
        {
            if ( !String.IsNullOrEmpty( pathName ) )
                UIPath = pathName;
            if ( !String.IsNullOrEmpty( fileName ) )
                UIFilename = fileName;

            string fn = Path.Combine( UIPath, UIFilename );
            Debug.WriteLine( "QRB saving UI code to " + fn );

            using ( StreamWriter writer = new StreamWriter( fn ) )
            {
                // Output header
                writer.WriteLine( "<?xml version=\"1.0\" encoding=\"utf-8\"?>" );
                writer.WriteLine( "<Page ButtonOpacityLevel=\"0.500\" Name=\"ProfitUI_QuickRaidButtons\">" );
                writer.WriteLine( "<Text DynamicData=\"/GameData.Self.ActualLevel\" Name=\"ActualLevel\">80</Text>" );
                writer.WriteLine( "" );

                foreach ( KeyValuePair<string, QRBButtonAssignment> kv in ButtonAssignments )
                {
                    QRBButtonAssignment ba = kv.Value;
                    QRBSpellClass sc = null;
                    if ( SpellClasses.ContainsKey( ba.Classname ) )
                        sc = SpellClasses[ba.Classname];
                    if ( sc == null )
                        continue;

                    // Output class Page
                    Debug.WriteLine( "QRB class block for " + ba.Classname );
                    writer.WriteLine( "<Page Name=\"" + ba.Classname + "\">" );

                    for ( int i = 1; i <= 5; i++ )
                    {
                        string sln = ba.getButton( i );
                        QRBSpellLine sl = null;
                        if ( !String.IsNullOrEmpty( sln ) )
                            sl = sc.findSpellLine( sln );
                        if ( sl != null )
                        {
                            Debug.WriteLine( "QRB button " + i.ToString() );
                            writer.Write( "<Data Name=\"Button" + i.ToString() + "\" Macro=\"" );
                            // Output level 0 literal commands first
                            foreach ( QRBSpell s in sl.Spells )
                            {
                                if ( s.Level == 0 )
                                {
                                    string o = doUISubstitutions( s.Text ).TrimStart( '/' ); // Remove leading slashes from commands if they included them
                                    writer.WriteLine( o );
                                }
                            }
                            // Then output spells in reverse order by level
                            List<QRBSpell> r = new List<QRBSpell>( sl.Spells );
                            r.Reverse();
                            bool haveSpells = false;
                            int iPrevLevel = 200; // Greater than max level for the forseeable future
                            foreach ( QRBSpell s in r )
                            {
                                if ( s.Level > 0 )
                                {
                                    string o = doUISubstitutions( s.Text );
                                    if ( !haveSpells )
                                    {
                                        // First spell is the highest level and is unconditional
                                        writer.WriteLine( "SpellForLevel=&quot;" + o + "&quot;" );
                                    }
                                    else
                                    {
                                        // After the first spell we output a conditional checking if our level is less
                                        // than the previous level and replacing the spell name if it is
                                        writer.WriteLine( "COND=(Parent.Parent.Parent.Parent.Custom.ProfitUI_QuickRaidButtons.ActualLevel.Text < " + iPrevLevel.ToString() + ")" );
                                        writer.WriteLine( "SpellForLevel=COND ? &quot;" + o + "&quot; : SpellForLevel" );
                                    }
                                    iPrevLevel = s.Level;
                                    haveSpells = true;
                                }
                            }
                            if ( haveSpells )
                                writer.Write( "useabilityonplayer Parent.Target SpellForLevel" );
                            writer.WriteLine( "\" Tooltip=\"" + sl.Tooltip + "\"/>" );
                        }
                        else
                        {
                            Debug.WriteLine( "QRB empty button " + i.ToString() );
                            writer.WriteLine( "<Data Name=\"Button" + i.ToString() + "\" Macro=\"NONE\" Tooltip=\"NO MACRO CONFIGURED\"/>" );
                        }
                    }

                    // Close class Page
                    writer.WriteLine( "</Page>" );
                    writer.WriteLine( "" );
                }

                // Output trailer
                writer.WriteLine( "</Page>" );
            }
        }
    }
}
