﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Diagnostics;

//  ProfitUI Quick Raid Buttons tool
//  Copyright (C) 2012 Todd T Knarr <tknarr@silverglass.org> <tknarr@cox.net> <todd.knarr@gmail.com>

//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.

//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.

//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.

namespace QuickRaidButtons
{
    /// <summary>
    /// When the Save button is pressed, store the internal spell classes object into the QRB storage
    /// object and save the QRB storage object to the storage XML file.
    /// When the Revert button is pressed, load the spell classes into the internal object from the QRB
    /// storage object.
    /// The Done and Cancel buttons act like Save and Revert respectively, and then return control to
    /// the button assignment form.
    /// </summary>
    public partial class SpellLines : Form
    {
        private static int spellNameCount = 16; // Recalculated in the constructor based on the number of spell name textboxes
        private string currentClassName = "";
        private string currentSpellLine = "";

        private QRBStorage qrb = null;
        private Dictionary<string, QRBSpellClass> spellClasses = null;

        public SpellLines( QRBStorage qrbs, string className )
        {
            InitializeComponent();

            qrb = qrbs;
            spellClasses = qrb.CloneSpellClasses();
            currentClassName = className;
            currentSpellLine = "";

            // Scan the controls array for our spell textboxes, and determine the maximum number of them
            int maxtb = 0;
            Control[] gbc = Controls.Find( "spells_groupBox", true );
            foreach ( Control ctrl in gbc[0].Controls )
            {
                if ( ( ctrl.Name.Length > 13 ) && ctrl.Name.StartsWith( "spell_textBox" ) )
                {
                    string num = ctrl.Name.Remove( 0, 13 );
                    int n = Int32.Parse( num ); // If this throws an exception, someone's messed up the source code
                    if ( n > maxtb )
                        maxtb = n;
                }
            }
            if ( maxtb > 0 )
                spellNameCount = maxtb;

            populateSpellLineFields();
        }

        /// <summary>
        /// Updates the internal structure for the current class based on the form field contents
        /// </summary>
        private void updateSpellLineData()
        {
            Debug.WriteLine( "Updating spell line data" );
            Debug.WriteLine( "Class: " + currentClassName );
            Debug.WriteLine( "Spell line: " + currentSpellLine );

            if ( String.IsNullOrEmpty( currentClassName ) || String.IsNullOrEmpty( currentSpellLine ) )
                return;

            // Locate our spell class object containing our spell lines, or create a new one
            // if one didn't already exist
            QRBSpellClass spellClass = null;
            if ( spellClasses.ContainsKey( currentClassName ) )
                spellClass = spellClasses[currentClassName];
            if ( spellClass == null )
            {
                Debug.WriteLine( "USL: created new spell class object " + currentClassName );
                spellClass = new QRBSpellClass( currentClassName );
                spellClasses[currentClassName] = spellClass;
            }

            // Locate the spell line we're editing. Create a new one if it doesn't
            // exist for this class
            QRBSpellLine spellLine = spellClass.findSpellLine( currentSpellLine );
            if ( spellLine == null )
            {
                Debug.WriteLine( "USL: created new spell line object " + currentSpellLine );
                spellLine = new QRBSpellLine( currentSpellLine, currentSpellLine );
                spellClass.SpellLines.Add( spellLine );
            }

            // Update the internal spell line object from the form fields
            spellLine.Tooltip = tooltip.Text;

            // Get spell line data from form fields
            spellLine.Spells.Clear();
            for ( int i = 1; i <= spellNameCount; i++ )
            {
                Control[] a_lvl = Controls.Find( "spell_numericUpDown" + i.ToString(), true );
                Control[] a_txt = Controls.Find( "spell_textBox" + i.ToString(), true );

                int lvl = 0;
                string txt = "";
                if ( a_txt.Length > 0 )
                {
                    TextBox c_txt = a_txt[0] as TextBox;
                    if ( c_txt != null )
                        txt = c_txt.Text;
                    if ( !String.IsNullOrEmpty( txt ) )
                    {
                        if ( a_lvl.Length > 0 )
                        {
                            NumericUpDown c_lvl = a_lvl[0] as NumericUpDown;
                            if ( c_lvl != null )
                            {
                                String v = c_lvl.Value.ToString();
                                if ( !String.IsNullOrEmpty( v ) )
                                    int.TryParse( v, out lvl );
                                else
                                    lvl = 0;
                            }
                        }
                        Debug.WriteLine( "USL: updating spell: " + lvl.ToString() + ": " + txt );
                        spellLine.addSpell( lvl, txt );
                    }
                }
            }
            Debug.WriteLine( "USL: resequencing spells" );
            spellLine.resequenceSpells();
        }

        /// <summary>
        /// Populates the spell line fields from the internal structure based on the current class
        /// </summary>
        private void populateSpellLineFields()
        {
            Debug.WriteLine( "Populating spell line fields" );
            Debug.WriteLine( "Class: " + currentClassName );
            Debug.WriteLine( "Spell line: " + currentSpellLine );

            // Locate our spell class object containing our spell lines, or create a new one
            // if one didn't already exist
            QRBSpellClass spellClass = null;
            if ( spellClasses.ContainsKey( currentClassName ) )
                spellClass = spellClasses[currentClassName];
            if ( spellClass == null )
            {
                Debug.WriteLine( "PSL: created new spell class object " + currentClassName );
                spellClass = new QRBSpellClass( currentClassName );
                spellClasses[currentClassName] = spellClass;
            }

            // If we don't have a current spell line name, use the name of the first spell line
            // for the class
            if ( string.IsNullOrEmpty( currentSpellLine ) && ( spellClass.SpellLines.Count > 0 ) && ( spellClass.SpellLines[0] != null ) )
                currentSpellLine = spellClass.SpellLines[0].Linename;

            // Reset the spell line form fields
            className.Text = currentClassName;
            spellLineName.ResetText();
            spellLineName.Items.Clear();
            tooltip.ResetText();
            Control[] gbc = Controls.Find( "spells_groupBox", true );
            foreach ( Control c in gbc[0].Controls )
            {
                if ( c.Name.StartsWith( "spell_textBox" ) )
                {
                    TextBox tb = c as TextBox;
                    tb.ResetText();
                }
                else if ( c.Name.StartsWith( "spell_numericUpDown" ) )
                {
                    NumericUpDown n = c as NumericUpDown;
                    n.ResetText();
                    n.Text = "0";
                    n.Value = 0;
                }
            }

            // If we don't have a spell line name, just use the blank form
            if ( string.IsNullOrEmpty( currentSpellLine ) )
                return;

            // Find our spell line object. If we can't find one, create a new empty one using
            // the current spell line name
            QRBSpellLine spellLine = spellClass.findSpellLine( currentSpellLine );
            if ( spellLine == null )
            {
                Debug.WriteLine( "PSL: created new spell line object " + currentSpellLine );
                spellLine = new QRBSpellLine( currentSpellLine, currentSpellLine );
                spellClass.SpellLines.Add( spellLine );
            }

            // Populate the controls based on the spell line object selected
            foreach ( QRBSpellLine sl in spellClass.SpellLines )
                spellLineName.Items.Add( sl.Linename );
            spellLineName.Text = spellLine.Linename;
            tooltip.Text = spellLine.Tooltip;
            int i = 1;
            foreach ( QRBSpell sp in spellLine.Spells )
            {
                string spinner_name = "spell_numericUpDown" + i.ToString();
                string tb_name = "spell_textBox" + i.ToString();

                Control[] spinners = Controls.Find( spinner_name, true );
                Control[] tbs = Controls.Find( tb_name, true );

                Debug.WriteLine( "PSL: updating spell " + i.ToString() + ": " + sp.Level.ToString() + ": " + sp.Text );
                if ( spinners.Length > 0 )
                {
                    NumericUpDown ud = spinners[0] as NumericUpDown;
                    ud.Text = sp.Level.ToString();
                    ud.Value = sp.Level;
                }
                if ( tbs.Length > 0 )
                    tbs[0].Text = sp.Text;

                i++;
                if ( i > spellNameCount )
                    break;
            }
        }

        private void doneButton_Click( object sender, EventArgs e )
        {
            Debug.WriteLine( "Done button clicked" );
            updateSpellLineData();
            qrb.SpellClasses = spellClasses;
        }

        private void cancelButton_Click( object sender, EventArgs e )
        {
            Debug.WriteLine( "Cancel button clicked" );
            spellClasses = qrb.CloneSpellClasses();
        }

        private void saveButton_Click( object sender, EventArgs e )
        {
            Debug.WriteLine( "Save button clicked" );
            updateSpellLineData();
            qrb.SpellClasses = spellClasses;
        }

        private void revertButton_Click( object sender, EventArgs e )
        {
            Debug.WriteLine( "Revert button clicked" );
            spellClasses = qrb.CloneSpellClasses();
            populateSpellLineFields();
        }

        private void className_SelectionChangeCommitted( object sender, EventArgs e )
        {
            Debug.WriteLine( "Class name selection changed" );
            ComboBox className = (ComboBox) sender;

            updateSpellLineData();
            currentClassName = className.Text;
            currentSpellLine = "";
            Debug.WriteLine( "New spell class selected: " + currentClassName );
            populateSpellLineFields();
        }

        private void spellLineName_SelectionChangeCommitted( object sender, EventArgs e )
        {
            Debug.WriteLine( "Spell line name selection changed" );
            ComboBox spellLineName = (ComboBox) sender;

            updateSpellLineData();
            currentSpellLine = spellLineName.Text;
            Debug.WriteLine( "New spell line selected: " + currentSpellLine );
            populateSpellLineFields();
        }

        private void clearButton_Click( object sender, EventArgs e )
        {
            Button btn = (Button) sender;
            Debug.WriteLine( "Clearing spell line fields - " + btn.Name );

            string levelSpinnerName = "spell_numericUpDown" + btn.Name.Remove( 0, 12 );
            string textBoxName = "spell_textBox" + btn.Name.Remove( 0, 12 );

            Control[] spinners = Controls.Find( levelSpinnerName, true );
            if ( spinners.Length > 0 )
            {
                NumericUpDown ud = spinners[0] as NumericUpDown;
                ud.ResetText();
                ud.Text = "0";
                ud.Value = 0;
            }
            Control[] tbs = Controls.Find( textBoxName, true );
            if ( tbs.Length > 0 )
                tbs[0].ResetText();
        }

        private void sortButton_Click( object sender, EventArgs e )
        {
            Debug.WriteLine( "Sort spells button clicked" );

            updateSpellLineData();
            populateSpellLineFields();
        }

        private void autoFillButton_Click( object sender, EventArgs e )
        {
            //string spellBaseText = "";
            //int spellLevel = 0;

            // TODO Auto-fill spell names based on the first spell name and it's level
        }

        private static string[] roman_numbers = { "", "I", "II", "III", "IV", "V", "VI", "VII", "VIII", "IX",
                                                  "X", "XI", "XII", "XIII", "XIV", "XV", "XVI", "XVII", "XVIII", "XIX" };

        /// <summary>
        ///  Convert an integer into a Roman numeric string
        ///  Note: this function only works on values in the 1-19 range
        /// </summary>
        /// <param name="i">Number to convert</param>
        /// <returns>Roman numeric form</returns>
        private string toRoman( int i )
        {
            return roman_numbers[i];
        }

        /// <summary>
        /// Convert a Roman numeric string to an int
        /// </summary>
        /// <param name="s">String to convert</param>
        /// <returns>Integer value</returns>
        private int fromRoman( string s )
        {
            for ( int i = 1; i < 20; i++ )
                if ( s.Equals( roman_numbers[i] ) )
                    return i;
            return 0;
        }

        private void deleteLineButton_Click( object sender, EventArgs e )
        {
            Debug.WriteLine( "Clear entire line button clicked" );
            Debug.WriteLine( "Class: " + currentClassName );
            Debug.WriteLine( "Spell line: " + currentSpellLine );

            QRBSpellClass spellClass = null;
            if ( spellClasses.ContainsKey( currentClassName ) )
                spellClass = spellClasses[currentClassName];
            if ( spellClass != null )
                spellClass.removeSpellLine( currentSpellLine );
            currentSpellLine = "";
            populateSpellLineFields();
            return;
        }

        private void newLineButton_Click( object sender, EventArgs e )
        {
            Debug.WriteLine( "New spell line button clicked" );
            Debug.WriteLine( "Class: " + currentClassName );

            updateSpellLineData();
            using ( NewSpellLine nsl = new NewSpellLine() )
            {
                DialogResult result = nsl.ShowDialog();
                if ( result == DialogResult.OK )
                {
                    string newSpellLine = nsl.SpellLineName;
                    QRBSpellClass spellClass = null;
                    if ( spellClasses.ContainsKey( currentClassName ) )
                        spellClass = spellClasses[currentClassName];
                    if ( spellClass != null )
                    {
                        QRBSpellLine sl = spellClass.findSpellLine( newSpellLine );
                        if ( sl == null )
                        {
                            Debug.WriteLine( "Adding new spell line: " + newSpellLine );
                            spellClass.addSpellLine( newSpellLine, newSpellLine );
                        }
                        currentSpellLine = newSpellLine;
                    }
                    populateSpellLineFields();
                }
            }
        }

        // When entering a textbox, select all text so that typing replaces the current contents
        private void textbox_EnterSelect( object sender, EventArgs e )
        {
            TextBox tb = sender as TextBox;
            if ( tb != null )
            {
                tb.SelectAll();
            }
        }

        // When entering a numeric up/down spinner, select all text so that typing replaces the current contents
        private void numericUpDown_EnterSelect( object sender, EventArgs e )
        {
            NumericUpDown ud = sender as NumericUpDown;
            if ( ud != null )
            {
                if ( String.IsNullOrEmpty( ud.Text ) )
                {
                    ud.ResetText();
                    ud.Text = "0";
                    ud.Value = 0;
                }
                ud.Select( 0, ud.Text.Length );
            }
        }
    }
}
